#include <lxpanel/plugin.h>
#include <stdio.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <linux/i2c.h>
#include <linux/i2c-dev.h>
#include <unistd.h>

#define I2C_SLAVEADDR_BQ2589x	(0x6a)
#define DEV_I2C "/dev/i2c-1"

uint8_t regread(int fd, uint8_t addr);
void regwrite(int fd, uint8_t addr, uint8_t val);

void set_oneshot_measurement(int fd)
{
  uint8_t val;

  val = regread(fd, 0x02);
  val &= ~(1 << 6);
  val |= 0x80;
  regwrite(fd, 0x02, val);
}

// デバイスからのレジスタ値の読み出し処理
uint8_t regread(int fd, uint8_t addr){
  union i2c_smbus_data data;
  struct i2c_smbus_ioctl_data args;
  int err;
  
  args.read_write = I2C_SMBUS_READ;
  args.command = addr;
  args.size = I2C_SMBUS_BYTE_DATA;
  args.data = &data;
  err = ioctl(fd, I2C_SMBUS, &args); // I2Cドライバへのリードのリクエストを発行
  if(err < 0){
    printf("ioctl error\n");
  }

  return data.byte; // リードした値を返す
}

// デバイスのレジスタへの書き込み処理
void regwrite(int fd, uint8_t addr, uint8_t val)
{
  union i2c_smbus_data data;
  struct i2c_smbus_ioctl_data args;
  int err;

  data.word = val;
  
  args.read_write = I2C_SMBUS_WRITE;
  args.command = addr;
  args.size = I2C_SMBUS_BYTE_DATA;
  args.data = &data;
  err = ioctl(fd, I2C_SMBUS, &args); // I2Cドライバへのライトのリクエストを発行
  if(err < 0){
    printf("ioctl error\n");
  }
}

// internal to the plugin source, not used by the 'priv' variable
static int iInstanceCount = 0;

/* プラグインのインスタンス */
typedef struct
{
  GtkWidget* plugin;
  LXPanel* panel;
  config_setting_t *settings;
  gint iMyId;
  int count;
  GtkWidget *pLabel;
  GtkWidget *icon;
  int i2c_fd;
  uint8_t reg_data[0x14+1];
} PluginInfo;

/* アイコンPNGを設定する */
static void set_icon(PluginInfo* c, char* filename)
{
  GdkPixbuf *pixbuf;
  char path[256];
  sprintf (path, "/home/pi/%s", filename);
  pixbuf = gdk_pixbuf_new_from_file_at_scale (path, 32, 32, TRUE, NULL);
  if (pixbuf != NULL){
    gtk_image_set_from_pixbuf (GTK_IMAGE (c->icon), pixbuf);
    g_object_unref (pixbuf);
  }
}

static const char* vbus_status[] = {
  "No Input",
  "USB Host SDP",
  "USB CDP(1.5A)",
  "USB DCP(3.25A)",
  "Adjustable High Voltage DCP (MacCharge)(1.5A)",
  "Unknown Adapter(500mA)",
  "Non-Standard Adapter(1A/2A/2.1A/2.4A)",
  "OTG"
};
static char* chrg_stat[] = {
  "Not Charging",
  "Pre-charge",
  "Fast Charging",
  "Charge Termination Done",
};
static char* pg_stat[] = {
  "Not Power Good",
  "Power Good"
};
char* sdp_stat[] = {
  "USB100 input is detected",
  "USB500 input is detected"
};
char* vsys_stat[] = {
  "Not in VSYSMIN regulation",
  "In VSYSMIN regulation"
};

/* 3段階バッテリーレベルの判定によるバッテリーレベル表示 */
static void set_icon_by_battery_voltage(PluginInfo* info, bool charging, float voltage)
{
  const float discharging_thresh[2] = {
    3.68f, // 3 -> 2
    3.44f  // 2 -> 1
  };
  const float charging_thresh[2] = {
    3.96f, // 1 -> 2
    4.10f  // 2 -> 3
  };
  int num_bars = 1;
  char fname_str[20];
  
  if (charging){
    if (voltage < charging_thresh[0]){
      num_bars = 1;
    } else if ((voltage >= charging_thresh[0]) && (voltage < charging_thresh[1])) {
      num_bars = 2;
    } else {
      num_bars = 3;
    }
  } else {
    if (voltage >= discharging_thresh[0]){
      num_bars = 3;
    } else if ((voltage >= discharging_thresh[1]) && (voltage < discharging_thresh[0])){
      num_bars = 2;
    } else {
      num_bars = 1;
    }
  }

  sprintf(fname_str,
	  "batt_%d%s.PNG",
	  num_bars,
	  charging ? "C" : "");

  set_icon(info, fname_str);
}

static gint timer_event(PluginInfo* info)
{
  char str[64];
  char tip_string[1024] = { 0 };
  int i;

  set_oneshot_measurement(info->i2c_fd);

  for(i=0;i<=0x14;i++){
    info->reg_data[i] = regread(info->i2c_fd, i);
  }

  /* ツールチップに設定する文字列の作成 */
  sprintf(str, " Battery voltage        : %1.3f[V]\n",2.304f + ((info->reg_data[0x0e] & 0x7f) * 0.02f));
  strcat(tip_string, str);
  sprintf(str, " VBUS Status            : %s\n", vbus_status[(info->reg_data[0x0b] & 0xe0) >> 5]);
  strcat(tip_string, str);
  sprintf(str, " Charging Status        : %s\n", chrg_stat[(info->reg_data[0x0b] & 0x18) >> 3]);
  strcat(tip_string, str);
  sprintf(str, " Power Good Status      : %s\n", pg_stat[(info->reg_data[0x0b] & 0x04) >> 2]);
  strcat(tip_string, str);
  sprintf(str, " USB Input Status       : %s\n", sdp_stat[(info->reg_data[0x0b] & 0x02) >> 1]);
  strcat(tip_string, str);
  sprintf(str, " VSYS Regulation Status : %s\n", vsys_stat[(info->reg_data[0x0b] & 0x01)]);
  strcat(tip_string, str);
  sprintf(str, " ADC Conversion os TS   : %1.3f[%%]\n", 21.0f + ((info->reg_data[0x10] & 0x07f) * (59.0f/127.0f)));
  strcat(tip_string, str);
  sprintf(str, " VBUS Good Status       : %s\n", (info->reg_data[0x11] & 0x80) ? "VBUS attached" : "Not VBUS attached");
  strcat(tip_string, str);
  sprintf(str, " ADC Conversiton of VBUS: %2.3f[V]\n", 2.60f + ((info->reg_data[0x11] & 0x7f) * ((15.3f-2.6f)/127.0f)));
  strcat(tip_string, str);
  sprintf(str, " Charge Current         : %d mA\n", (info->reg_data[0x12] & 0x7f) * 50);
  strcat(tip_string, str);

  /* ツールチップへの文字列の設定 */
  gtk_widget_set_tooltip_text(info->icon, tip_string);

  /* バッテリー残量状態に応じて対応するピクトグラムアイコンを設定する */
  set_icon_by_battery_voltage(info,
			      ((info->reg_data[0x0b] & 0x18) >> 3) ? true : false,
			      2.304f + ((info->reg_data[0x0e] & 0x7f) * 0.02f));
  
  return TRUE;
}


GtkWidget *batt_constructor(LXPanel *panel, config_setting_t *settings)
{
  PluginInfo *info = g_new0(PluginInfo, 1);

  info->iMyId = ++iInstanceCount;

  info->icon = gtk_image_new();
  set_icon(info, "batt_1.PNG");
  gtk_widget_set_tooltip_text(info->icon, "Charger information");
  gtk_widget_set_visible(info->icon, TRUE);

  info->panel = panel;
  info->plugin = gtk_button_new ();
  gtk_button_set_relief (GTK_BUTTON (info->plugin), GTK_RELIEF_NONE);
  info->settings = settings;

  gtk_container_add (GTK_CONTAINER (info->plugin), info->icon);

  /* ウィジェットの表示 */
  gtk_widget_show_all (info->plugin);

  gtk_widget_set_size_request(info->plugin, 40, 25);

  g_timeout_add(2000, (GSourceFunc)timer_event, (gpointer)info);

  info->i2c_fd = open(DEV_I2C, O_RDWR);
  if(info->i2c_fd < 0){
    printf("i2c open error\n");
    exit(EXIT_FAILURE);
  }

  // I2Cスレーブアドレスをデバイスドライバに設定する
  ioctl(info->i2c_fd, I2C_SLAVE, I2C_SLAVEADDR_BQ2589x);

  {
    uint8_t val;

    val = regread(info->i2c_fd, 0x00);
    regwrite(info->i2c_fd, 0x00, val & (~(1 << 6)));

    val = regread(info->i2c_fd, 0x02);
    regwrite(info->i2c_fd, 0x02, val & (~(1 << 4)));

    val = regread(info->i2c_fd, 0x07);
    regwrite(info->i2c_fd, 0x07, val & (~(3 << 4)));
  }
  
  return info->plugin;
}

FM_DEFINE_MODULE(lxpanel_gtk, test)

/* プラグインの定義 */
LXPanelPluginInit fm_module_init_lxpanel_gtk = {
  .name = "Battery Monitor(BQ2589x)", /* プラグイン名称 */
  .description = "Battery monitor for RaspberryPi with Ti BQ2589x",

  // コンストラクタ関数の設定
  .new_instance = batt_constructor
};
