#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <linux/i2c.h>
#include <linux/i2c-dev.h>
#include <unistd.h>

#define I2C_SLAVEADDR_BQ2589x	(0x6a) // I2Cスレーブアドレス

uint8_t regread(int fd, uint8_t addr);
void regwrite(int fd, uint8_t addr, uint8_t val);

/*
  チャージャーICのADC変換開始を指示する
*/
void set_oneshot_measurement(int fd)
{
  uint8_t val;

  val = regread(fd, 0x02); // REG02の値を読み出し
  val &= ~(1 << 6); // One shot ADC conversionを設定
  val |= 0x80; // Start ADC Conversionを設定
  regwrite(fd, 0x02, val); // チャージャーICに値を書き込み
}

/*
  ツールの使用方法を表示
 */
void usage(void)
{
  printf("usage: bq25895test_3 [option]\n");
  printf("  -v num   : Specify charge voltage limit\n");
  printf("             1, 0x2A (4.000V)\n");
  printf("             2, 0x42 (4.096V) default\n");
  printf("             3, 0x5E (4.208V)\n");
  printf("  -c num   : Specify fast charge current limit\n");
  printf("             1, 0x10 (1024mA)\n");
  printf("             2, 0x20 (2048mA) default\n");
}

int main(int argc, char** argv)
{
  int fd;
  int i;
  uint8_t reg[0x14+1];
  uint8_t charge_voltage = 2;
  uint8_t charge_current = 2;

  /* コマンドラインオプションのチェック */
  {
    int ch;
    extern char* optarg;
    extern int optind, opterr;
    while ((ch = getopt(argc, argv, "v:c:h")) != -1){
      switch(ch){
      case 'v':
	charge_voltage = atoi(optarg);
	printf("Charge voltage specified : %d\n", charge_voltage);
	break;
      case 'c':
	charge_current = atoi(optarg);
	printf("Charge current specified : %d\n", charge_current);
	break;
      case 'h':
      default:
	usage();
	return 1;
	break;
      }
    }
  }

  char* vbus_status[] = {
    "No Input",
    "USB Host SDP",
    "USB CDP(1.5A)",
    "USB DCP(3.25A)",
    "Adjustable High Voltage DCP (MaxCharge)(1.5A)",
    "Unknown Adapter(500mA)",
    "Non-Standard Adapter(1A/2A/2.1A/2.4A)",
    "OTG"
  };
  char* chrg_stat[] = {
    "Not Charging",
    "Pre-charge",
    "Fast Charging",
    "Charge Termination Done",
  };
  char* pg_stat[] = {
    "Not Power Good",
    "Power Good"
  };
  char* sdp_stat[] = {
    "USB100 input is detected",
    "USB500 input is detected"
  };
  char* vsys_stat[] = {
    "Not in VSYSMIN regulation",
    "In VSYSMIN regulation"
  };

  fd = open("/dev/i2c-1",O_RDWR); // I2Cのデバイスドライバのオープン
  if(fd < 0){
    printf("i2c open error\n");
    exit(EXIT_FAILURE); // オープンできなかった場合には終了コード=異常をセットして終了
  }

  // I2Cスレーブアドレスをデバイスドライバに設定する
  ioctl(fd, I2C_SLAVE, I2C_SLAVEADDR_BQ2589x);

  {
    // 充電電圧の設定
    uint8_t val;
    switch(charge_voltage){
    case 1:
      val = 0x2a;
      break;
    case 2:
      val = 0x42;
      break;
    case 3:
      val = 0x5e;
      break;
    default:
      printf("Voltage limit error\n");
      return -1;
    }
    regwrite(fd, 0x06, val);

    // 充電電流の設定
    switch(charge_current){
    case 1:
      val = 0x10;
      break;
    case 2:
      val = 0x20;
      break;
    default:
      printf("Current limit error\n");
      return -1;
    }
    regwrite(fd, 0x04, val);
  }
  
  while(1){
    set_oneshot_measurement(fd); // ADC測定を開始させる
    sleep(1); // 測定完了待ち
    
    printf("\n--------------------------------------------------\n");
    // REG00～REG14の全値を読み出して表示
    printf("REG00-14:");
    for(i=0;i<=0x14;i++){
      reg[i] = regread(fd, i);
      printf("%02Xh,", reg[i]);
    }
    printf("\n");

    // 一部のレジスタ値を見やすく表示
    printf("REG0E = %02Xh\n", reg[0x0e]);
    printf(" Battery voltage        : %1.3f[V]\n",2.304f + ((reg[0x0e] & 0x7f) * 0.02f));
    
    printf("REG0B = %02Xh\n", reg[0x0b]);
    printf(" VBUS Status            : %s\n", vbus_status[(reg[0x0b] & 0xe0) >> 5]);
    printf(" Charging Status        : %s\n", chrg_stat[(reg[0x0b] & 0x18) >> 3]);
    printf(" Power Good Status      : %s\n", pg_stat[(reg[0x0b] & 0x04) >> 2]);
    printf(" USB Input Status       : %s\n", sdp_stat[(reg[0x0b] & 0x02) >> 1]);
    printf(" VSYS Regulation Status : %s\n", vsys_stat[(reg[0x0b] & 0x01)]);

    printf("REG10 = %02Xh\n", reg[0x10]);
    printf(" ADC Conversion of TS   : %1.3f[%%]\n", 21.0f + ((reg[0x10] & 0x07f) * (59.0f/127.0f)));
    
    printf("REG11 = %02Xh\n", reg[0x11]);
    printf(" VBUS Good Status       : %s\n", (reg[0x11] & 0x80) ? "VBUS attached" : "Not VBUS attached");
    printf(" ADC Conversiton of VBUS: %2.3f[V]\n", 2.60f + ((reg[0x11] & 0x7f) * ((15.3f-2.6f)/127.0f)));
    
    printf("REG12 = %02Xh\n", reg[0x12]);
    printf(" Charge Current         : %d mA\n", (reg[0x12] & 0x7f) * 50);

  }
}

// デバイスからのレジスタ値の読み出し処理
uint8_t regread(int fd, uint8_t addr){
  union i2c_smbus_data data;
  struct i2c_smbus_ioctl_data args;
  int err;
  
  args.read_write = I2C_SMBUS_READ;
  args.command = addr;
  args.size = I2C_SMBUS_BYTE_DATA;
  args.data = &data;
  err = ioctl(fd, I2C_SMBUS, &args); // I2Cドライバへのリードのリクエストを発行
  if(err < 0){
    printf("ioctl error\n");
  }

  return data.byte; // リードした値を返す
}

// デバイスのレジスタへの書き込み処理
void regwrite(int fd, uint8_t addr, uint8_t val)
{
  union i2c_smbus_data data;
  struct i2c_smbus_ioctl_data args;
  int err;

  data.word = val;
  
  args.read_write = I2C_SMBUS_WRITE;
  args.command = addr;
  args.size = I2C_SMBUS_BYTE_DATA;
  args.data = &data;
  err = ioctl(fd, I2C_SMBUS, &args); // I2Cドライバへのライトのリクエストを発行
  if(err < 0){
    printf("ioctl error\n");
  }
}

